//	Last Updated:	06/01/2005
//
//	xICE - Secure Internet Communications Encryption
//	Copyright (C) 2002-2005 Nick Hauenstein.
//	This program is distributed under the terms of the GNU
//	General Public License.
//
//      This file is part of the xICE Software Development Kit.
//
//	The xICE Software Development Kit is free software; you can
//	redistribute it and/or modify it under the terms of the GNU
//	General Public License as published by the Free Software
//	Foundation; either version 2 of the License, or (at your
//	option) any later version.
//	
//	The xICE Software Development Kit is distributed in the
//	hope that it will be useful, but WITHOUT ANY WARRANTY;
//	without even the implied warranty of MERCHANTABILITY or
//	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
//	Public License for more details.
//
//	You should have received a copy of the GNU General Public
//	License along with the xICE Software Development Kit; if
//	not, write to the Free Software Foundation, Inc.,
//	51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

//  Note: SOme minir changes to the orignal were amde by James Britt
//  as part of inclusion in de.linque.nt.  All still GPL and stuff.
//
//	------------------------------------------------------------------
//
//	This is a JavaScript implentation of the xICE string encryption
//	and decryption functions based off the original Visual Basic 6.0
//	code. Use this as a guide for future ports to other languages.
//	I went ahead an ported the Hex Conversion functions to get the
//	hang of the JavaScript language. This may actually be necessary
//	to do in some languages, as hex functions differ in how they return
//	results. In the original Visual Basic 6.0 code I wrote them because
//	the VB Hex function didn't produce results with exactly two
//	characters. My version of the functions produces exactl two whether
//	or not two are needed to express the value. Re-writing the functions
//	from scratch was a little harder than just writing a wrapper
//	for the ones in the VB runtime, but overall the code executes
//	faster. May not be the case for all languages, but do research it
//	before you waste any time.
//
//	Here's some last minute notes on the same subject (porting the
//	code). When creating an xICE port, be sure to consider carefully
//	the following:
//
//	1.) Precision to which Pi is represented... Pi in xICE is defined
//	    as 3.14159265358979
//	2.) Precision to which Sin & Cos are calculated. There should be
//	    exactly 15 digits trailing the decimal point.
//	3.) How the MakeRandom function rounds off its results. It should
//	    work as much like VB's INT function as possible.
//	
//	All of these issues pertain to the PRNG portion of the code. If
//	that portion is not produced to these specifications, encrypted
//	data produced by such a ported version may not be able to be
//	decrypted by other implementations of the xICE algorithm. Do not
//	settle for results coming out even .000000000000001 off, over long
//	strings of data, such a difference could have some disasterous
//	consequences.
//
//
//	What's available in this file?
//
//	Globals:
//
//	dblCenterY - Y-value for the center of the circle in the PRNG. Used
//		internally.
//	dblCenterX - X-value for the center of the circle in the PRNG. Used
//		internally.
//	LastResult - Holds the last result of EncryptText & DecryptText
//		functions.
//	LastErrNum - Holds the last error number generated by either the
//		EncryptText or the DecryptText function.
//	LastErrDes - Hold the description of the last error generated by
//		either the EncryptText or the DecryptText function.
//
//	Functions:
//
//	Hex Conversion Functions:
//	HexToNum(strInput) - Converts a two-digit hex code into a byte value.
//		Used internally.
//	NumToHex(intNum) - Converts a byte value into a two-digit hex code.
//		Used internally.
//
//	Randomization Functions:
//	ReCenter(mdblCenterY,mdblCenterX) - ReCentes the circle in the PRNG
//		to the coordinates (mdblCenterX, mdblCenterY). Used internally.
//	Generate(dblRadius,dblTheta) - Given a Radius (dblRadius) and an
//		angle (dblTheta), generates new X and Y values for the PRNG.
//		Used internally.
//	MakeRandom(dblValue,lngMax) - Converts a random number (dblValue) into
//		a number between 0 and lngMax. Used internally.
//	
//	Cryptographic Functions:
//	GetStrength(strPassword) - Returns a number representing the strength
//		of a password (strPassword) in bits.
//	GetSerial() - Returns the Time Serial to be used in encryption. Used
//		internally.
//	EncryptText(strClear, strKey) - Encrypts clear text (strClear) by the
//		Key (strKey), and stores the result in LastResult. Returns 1
//		if the operation was a success, and -1 if the operation failed.
//	DecryptText(strCipher, strKey) - Decrypts cipher text (strCipher) using
//		the Key (strKey), and stores the result in LastResult. Returns 1
//		if the operation was a success, and -1 if the operation failed.	
//
//	Error Handling Functions:
//	RaiseError(intErrNum, strErrDes) - Stores intErrNum as the value
//	for LastErrNum, and strErrDes as the value for LastErrDes. Used
//	internally.
//
// ********************				Hex Conversion Functions
function HexToNum(strInput) {
	
	// This is just in case some parts of the encrypted data went down
	// to lowercase, or were typed in as lowercase.
	strInput.toUpperCase();
	
	// This function is pretty self-explanatory. Just a simple Hex 
	// conversion function.
	
	var intSixteens = 0;
	var intOnes = 0;

	switch (strInput.substr(0,1)) {
	
		case "0":
			intSixteens = parseInt(strInput.substr(0,1));
			break;	
		case "1":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "2":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "3":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "4":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "5":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "6":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "7":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "8":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "9":
			intSixteens = parseInt(strInput.substr(0,1));		
			break;
		case "A":
			intSixteens = 10;
			break;
		case "B":
			intSixteens = 11;
			break;
		case "C":
			intSixteens = 12;
			break;
		case "D":
			intSixteens = 13;
			break;
		case "E":
			intSixteens = 14;
			break;
		case "F":
			intSixteens = 15;
			break;
	}

	switch (strInput.substr(1,1)) {
	
		case "0":
			intOnes = parseInt(strInput.substr(1,1));	
			break;
		case "1":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "2":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "3":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "4":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "5":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "6":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "7":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "8":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "9":
			intOnes = parseInt(strInput.substr(1,1));		
			break;
		case "A":
			intOnes = 10;
			break;
		case "B":
			intOnes = 11;
			break;
		case "C":
			intOnes = 12;
			break;
		case "D":
			intOnes = 13;
			break;
		case "E":
			intOnes = 14;
			break;
		case "F":
			intOnes = 15;
			break;
	}

	return (intSixteens * 16) + intOnes;
}

function NumToHex(intNum) {

	// This function converts a decimal number between 1 and 255 to
	// Hex. This is another self-explanatory one.

	var strOnes = "";
	var intOnes;

	var strSixteens = "";
	var intSixteens;
	
	intSixteens = intNum / 16;
	intSixteens = Math.floor(intSixteens);

	intOnes = intNum - (16 * intSixteens);

	switch (intOnes) {

		case 10:
			strOnes = "A";
			break;
		case 11:
			strOnes = "B";
			break;
		case 12:
			strOnes = "C";
			break;
		case 13:
			strOnes = "D";
			break;
		case 14:
			strOnes = "E";
			break;
		case 15:
			strOnes = "F";
			break;
		default:
			strOnes = intOnes.toString();
			break;
	}	

	switch (intSixteens) {

		case 10:
			strSixteens = "A";
			break;
		case 11:
			strSixteens = "B";
			break;
		case 12:
			strSixteens = "C";
			break;
		case 13:
			strSixteens = "D";
			break;
		case 14:
			strSixteens = "E";
			break;
		case 15:
			strSixteens = "F";
			break;
		default:
			strSixteens = intSixteens.toString();
			break;
	}	

	return strSixteens + strOnes;

}
// ********************				Randomization Functions
var dblCenterY;
var dblCenterX;

function ReCenter(mdblCenterY,mdblCenterX) {

	// This just allows the Generate function to change up some
	// global variables. It could do it on its own, but putting it
	// here makes the code down there read better without heavily
	// commenting. Even though it already is heavily commented.

	dblCenterY = mdblCenterY;
	dblCenterX = mdblCenterX;
	
}

function Generate(dblRadius,dblTheta) {

	// It is very important that these numbers are EXACTLY the same
	// in ALL implementations to allow universal compatibility.

	// MAGIC NUMBERS
	// KEEP PI TO THIS PRECISION, NO MORE, NO LESS
	var sngPi = 3.14159265358979;
	
	// Boundaries of our nearly perfect square coordinate system...
	var sngMaxUpper = 2147483647;
	var sngMaxLower = -2147483648;

	// Basically what we're doing here is picking a point on a circle
	// centered at (dblCenterX, dblCenterY). This point will serve as
	// the center of the next circle used for this function. The
	// radius of the circle is given to the function, as well as the
	// angle the new point makes with the center of the orignal circle.
	// This angle and radius is ultimately what determines the new
	// point, and serve as our pseudo-random seeds

	var dblResultX;
	var dblResultY;

	dblResultX = (dblRadius * Math.cos((dblTheta / 180) * sngPi).toFixed(15)) + dblCenterX;

	dblResultY = (dblRadius * Math.sin((dblTheta / 180) * sngPi).toFixed(15)) + dblCenterY;

	if (dblResultX > sngMaxUpper || dblResultX < sngMaxLower) {

		// Re-center if new X coordinate is far off the boundary of the X-axis
		ReCenter(dblCenterY, 0);

	} else {

		// Otherwise, calculate the new X-coordinate
		dblCenterX = dblResultX;

	}

	if (dblResultY > sngMaxUpper || dblResultY < sngMaxLower) {

		// Re-center if new Y coordinate is far off the boundary of the Y-axis
		ReCenter(0,dblCenterX);		

	} else {

		// Otherwise, calculate the new Y-coordinate
		dblCenterY = dblResultY;

	}

}

function MakeRandom(dblValue,lngMax) {

	// This is a cheap way of taking an arbitrary value and using it 
	// to pick a number between 0 and lngMax. This is heavily modified
	// from the Visual Basic version. Mainly because VB handles this a lot
	// more gracefully, thanks to it's crazy Mod operator. Which in truth,
	// I dislike greatly for its handling of large numbers, or lack
	// of handling... whichever is closer to the truth. (The crazy mod
	// operator is why the boundaries of the coordinate system in the
	// PRNG had to be limited to begin with.
	
	var intRandom;

	intRandom = (dblValue % (lngMax + 1)).toFixed(0);

	intRandom = Math.abs(intRandom);
	
	if (intRandom > lngMax) {

		intRandom = 0;

	}

	return intRandom;

}
// ********************				Cryptographic Functions
var LastResult;
var LastErrNum;
var LastErrDes;

function GetStrength(strPassword) {

	var strSerial;
	
	// Get the time serial
	strSerial = GetSerial();

	// Multiply the password's length by 8 (8 bits per byte, 1 byte per character,
	// 8*1 = 8, and the same for the time serial's length, which for the purposes
	// of the xICE algorithm is considered part of the key.
	return (strPassword.length * 8) + (strSerial.length * 8);

}

function GetSerial() {
	// Basically this takes a date object, and makes a lovely string of it.
	// For this implementation, this serves as the time data. However,
	// I'd recommend using something that isn't as easy to figure out.
	// one weakness of the xICE algorithm is that knowing the time data
	// or even a part of it could reveal part of the key. Assuming you
	// find where the time data is within the cipher text. It's highly
	// unlikely, but still do try and protect against it.
	var dtTimeSerial = new Date();
	var strTimeSerial

	strTimeSerial = dtTimeSerial.toGMTString();

	strTimeSerial = strTimeSerial.toString();

	return strTimeSerial;
}

function EncryptText(strClear, strKey) {

	// Error Constants. These will be stored
	// in LastErrNum if there is a problem.
	var errInvalidKeyLength = 65101;
	var errInvalidSize = 65103;
	var errKeyMissing = 65303;
	var errClearTextMissing = 65305;

	// Declare variables to be used...
	var strCipher;
	var arbyClear = new Array(strClear.length);
	var arbyKey = new Array(strKey.length);
	var strTimeSerial;
	var intTimeSerialLength;
	var strTimeSerialLength;
	var intSerialPosition;
	var intPosition;
	var blnSecondValue;
	var i;
	var k;

	// Do some validation work, bail if there's a problem.
	if (strClear == "") {
	
		RaiseError(errClearTextMissing, "Clear Text Missing");
		return -1;

	}

	if (strKey == "") {
	
		RaiseError(errKeyMissing, "Key Missing");
		return -1;
	
	}
	
	if (strKey.length <= 1) {
	
		RaiseError(errInvalidKeyLength, "Invalid Key Length");
		return -1;
		
	}
	
	// This generates a time serial. This can really be changed
	// into any data that you see fit. As long as it is not easily
	// determined, it is kind of like a secondary key, as it is
	// also used to seed the PRNG

	strTimeSerial = GetSerial();

	intTimeSerialLength = strTimeSerial.length;

	// XOR the byte that holds the length of the time serial by the
	// first byte of the key string

	strTimeSerialLength = NumToHex(intTimeSerialLength ^ strKey.charCodeAt(0));

	var arbyTimeSerial = new Array(strTimeSerial.length);

	// These two for loops load up our strings into arrays so they
	// can be easily walked through later without having to use
	// functions that have names that I will forget within a few
	// seconds of hearing them (charAt... What is that?!). Anyhoo
	// it makes everything look clean at least.

	if (strClear.length == 0) {
	
		RaiseError(errInvalidSize, "Text Has Zero Length");
		return -1;
		
	}
	
	for (i = 0; i < strClear.length; i++) {

		arbyClear[i + 1] = strClear.charCodeAt(i);

	}

	for (i = 0; i < strKey.length; i++) {

		arbyKey[i + 1] = strKey.charCodeAt(i);

	}	


	for (i = 0; i < intTimeSerialLength; i++) {

		arbyTimeSerial[i + 1] = strTimeSerial.charCodeAt(i);

	}	

	// This code creates a hash of sorts of the key to be used later
	// to verify if the correct password was used in the decrypt
	// function. The process is to basically xor the first letter of
	// the password with the last letter, the second with the
	// second to the last etc... If there are an odd number of
	// characters in the password, the very middle character is
	// ignored.

	strCipher = "";
	intPosition = 2;

	for (k = 1; k < Math.floor(strKey.length / 2) + 1; k++) {

		// This code is a bugger, and different than the VB code in that
		// JScript strings are 0 based... hence all the - k + 1 garbage...

		strCipher = strCipher.concat(NumToHex(arbyKey[k] ^ arbyKey[(strKey.length - k) + 1]));

	}

	// Okay, now that the key hash is stored, it's time to store the time serial
	// length, followed by the time serial itsef.

	strCipher = strCipher.concat(strTimeSerialLength);

	// The length byte and the time serial are both XORed by the key
	// in a straight up block cipher with a block length equal to the key
	// length. You'll figure it out, look at the code, it is very simple.
	
	for (i=1; i < (strTimeSerial.length + 1); i++) {

		strCipher = strCipher.concat(NumToHex(arbyKey[intPosition] ^ arbyTimeSerial[i]));

		if (intPosition == strKey.length) {
	
			intPosition = 1;		

		} else {

			intPosition++;
		
		}
		
	} 
	
	intPosition = 1;

	intSerialPosition = 1;

	// This seeds the random generator with a better starting point on the graph
	// than (0,0), because that would be pretty darn predictable, and could reveal
	// other stuff down the road.
	
	ReCenter(arbyKey[intPosition],arbyTimeSerial[intSerialPosition]);

	intPosition++;

	intSerialPosition++;

	// This is the actual encryption operation, pay attention...

	// This variable keeps track of whether to use the X-value generated, or Y-value.
	// each time the PRNG is run, it is seeded with 2 numbers, and results in 2 numbers,
	// It is seeded everytime the first value is used.
	
	blnSecondValue = -1;

	for (i = 1; i < (strClear.length + 1); i++) {

		if (blnSecondValue == -1) {

			// Generate a new Circle Center every other time, use the X first

			Generate(arbyKey[intPosition], arbyTimeSerial[intSerialPosition]);

			// Here's where every odd-numbered byte of the cleartext is encrypted
			strCipher = strCipher.concat(NumToHex(arbyClear[i] ^ MakeRandom(dblCenterX, 255)));

			blnSecondValue = 1;

		} else {

			// Then use the Y...
			// Here's where every even-numbered byte of the cleartext is encrypted
			strCipher = strCipher.concat(NumToHex(arbyClear[i] ^ MakeRandom(dblCenterY, 255)));

			blnSecondValue = -1;

		}

		// Keep track of where we are at in walking through the key
		if (intPosition == strKey.length) {

			intPosition = 1;
	
		} else {

			intPosition++;

		}
		
		// Keep track of where we are at in walking through the time serial
		if (intSerialPosition == strTimeSerial.length) {

			intSerialPosition = 1;

		} else {

			intSerialPosition++;

		}

	}

	LastResult = strCipher;
	
	return 1;

}

function DecryptText(strCipher, strKey) {

	// Error constants to be yelled at the user...
	var errInvalidKeyLength = 65101;
	var errInvalidSize = 65103;
	var errInvalidKey = 65102;
	var errKeyMissing = 65303;
	var errCipherTextMissing = 65304;

	// Variables we'll be using...
	var arbyCipher;
	var arbyTimeSerial;
	var arbyKey = new Array(strKey.length);
	var intTimeSerialLength;
	var intCipherTextLength;
	var strCheckHash = "";
	var intSerialPosition;
	var intPosition;
	var blnSecondValue;
	var strClear = "";
	var i;
	var k;

	// Some error checking, before we even get started...
	if (strCipher == "") {
	
		RaiseError(errCipherTextMissing, "Cipher Text Missing");
		return -1;
	
	}

	if (strCipher.length < 10) {
	
		RaiseError(errInvalidSize, "Bad Text Length");
		return -1;
		
	}
	
	if (strKey == "") {
		
		RaiseError(errKeyMissing, "Key Missing");
		return -1;
		
	}
	
	if (strKey.length <=1) {
		
		RaiseError(errInvalidKeyLength, "Invalid Key Length");
		return -1;
		
	}
	
	// Loading up the key into a byte array, for easier access...
	for (i = 0; i < strKey.length; i++) {

		arbyKey[i + 1] = strKey.charCodeAt(i);

	}	
	
	for (k = 1; k < Math.floor(strKey.length / 2) + 1; k++) {

		// This code is still a bugger, and different than the VB code in that
		// JScript strings are 0 based... hence all the - 1 + 1 garbage...

		// Here we're making up the hash of the key again, to check it against
		// the one in the cipher text. If it matches, it's likely that the user typed
		// the right key, and it's okay to proceed. If it doesn't match, it's the wrong
		// password, or there's a bug in the code somewhere. Speaking of which, here's a
		// quick but extremely important note: If you're making a port and want to know
		// if your Encrypt function is working, test it with the Decrypt function of
		// ANOTHER implementation.
		strCheckHash = strCheckHash.concat(NumToHex(arbyKey[k] ^ arbyKey[(strKey.length - k) + 1]));

	}
	
	// Here's where we check if the password is wrong, or if it's right.
	if (strCheckHash != strCipher.substring(0, strCheckHash.length)) {
	
		RaiseError(errInvalidKey, "Invalid Key");
		return -1;
	
	}
	
	// Now we cut off the hash from the front of the cipher text, because
	// we don't need it anymore.
	strCipher = strCipher.substring(strCheckHash.length, strCipher.length + 1);
	
	// And here we find out the length of the Time Serial by first XORing the
	// next byte in the file by the first byte in the key.
	intTimeSerialLength = arbyKey[1] ^ HexToNum(strCipher.substring(0, 2));
	
	// And we cut that byte out after we're done with it.
	strCipher = strCipher.substring(2, strCipher.length + 1);
	
	// And we make sure that we remember what position of the key we're at.
	intPosition = 2;
	
	// Make the array for the time serial. Again stored in an array because
	// it's easier that way, and makes the code pretty.
	arbyTimeSerial = new Array(intTimeSerialLength);
	
	for (i = 1; i <= intTimeSerialLength; i++) {
	
		arbyTimeSerial[i] = arbyKey[intPosition] ^ HexToNum(strCipher.substring(0, 2));

		// Cut each byte off after it's decrypted and put into the array.
		strCipher = strCipher.substring(2, strCipher.length + 1);
	
		// Keep track of where we are at in walking through the key
		if (intPosition == strKey.length) {

			intPosition = 1;
	
		} else {

			intPosition++;

		}
		
	}
	
	// We're going to reset to the beginning of the key, and time serial now.
	intPosition = 1;

	intSerialPosition = 1;

	// Now we seed up the random generator... Getting prepared to decrypt
	
	ReCenter(arbyKey[intPosition],arbyTimeSerial[intSerialPosition]);

	intPosition++;

	intSerialPosition++;

	blnSecondValue = -1;
	
	// Load strCipher into array, at the last minute...
	
	// First we need to remember how long it was...
	intCipherTextLength = strCipher.length / 2;
	
	// Now we build our array...
	arbyCipher = new Array(intCipherTextLength);
	
	for (i = 1; i <= intCipherTextLength; i++) {

		// Fill up array with translated ciphertext...
		arbyCipher[i] = HexToNum(strCipher.substring(0, 2));

		// Removes data as its translated, saves memory and makes
		// for easier to read code.
		strCipher = strCipher.substring(2, strCipher.length + 1);
	
	}
	
	// Here's the decryption routine...
	
	for (i = 1; i <= intCipherTextLength; i++) {

		if (blnSecondValue == -1) {

			// Generate a new Circle Center every other time, use the X first

			Generate(arbyKey[intPosition], arbyTimeSerial[intSerialPosition]);

			strClear = strClear.concat(String.fromCharCode(arbyCipher[i] ^ MakeRandom(dblCenterX, 255)));

			blnSecondValue = 1;

		} else {

			// Then use the Y...

			strClear = strClear.concat(String.fromCharCode(arbyCipher[i] ^ MakeRandom(dblCenterY, 255)));

			blnSecondValue = -1;

		}

		// Keep track of where we are at in walking through the key
		if (intPosition == strKey.length) {

			intPosition = 1;
	
		} else {

			intPosition++;

		}
		
		// Keep track of where we are at in walking through the time serial
		if (intSerialPosition == intTimeSerialLength) {

			intSerialPosition = 1;

		} else {

			intSerialPosition++;

		}

	
	}
	
	LastResult = strClear;
	
	return 1;
}

function RaiseError(intErrNum, strErrDes) {

	LastErrNum = intErrNum;
	LastErrDes = strErrDes
	
	return;
}  
//SPLITME
//   var key = 'SEEKRET' // for xICE, but use your own key!
//   el.innerHTML = DecryptText( el.innerHTML, key ) //for xIce
//SPLITME
//   var key = 'SEEKRET' // for xICE, but use your own key!
//   el.innerHTML = DecryptText( el.innerHTML, key ) //for xIce
